/***********************************************************************/
/* FreeVR OpenGL tutorial example 7 (for C++): ex7_travel.C            */
/* Last tested with: FreeVR version 0.5g                               */
/* Last Modified: May 2, 2005                                          */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/* C++ Version                                                         */
/*                                                                     */
/* This example code demonstrates a VR application program that        */
/*   allows the user to fly the world using a joystick controlled      */
/*   point and fly method.                                             */
/*                                                                     */
/*                                                                     */
/* The new features of this example include:                           */
/*   - FreeVR Library functions, variables, and macros:                */
/*	* vrGetValuatorValue() [used for joystick interaction]         */
/*      * vrPointGetVWFromUser6sensor()    [versus the real-world]     */
/*      * vrVectorGetRWFrom6sensorDir()                                */
/*      * vrUserTravelReset()                                          */
/*      * vrUserTravelRotateId()                                       */
/*      * vrUserTravelTranslate3d()                                    */
/*      * vrRenderTransformUserTravel(rendinfo)                        */
/*   - Objects fixed with respect to the real world vs. virtual world  */
/*      (ie. coordinate systems)                                       */
/*                                                                     */
/* Considerations:                                                     */
/*   - A floor/ground helps user to maintain bearings.                 */
/*   - Now concerned about how much time passes between sim frames.    */
/*   - Interactions between wand and objects must take place in the    */
/*      same coordinate system reference frame.                        */
/*                                                                     */
/* This application uses the files:                                    */
/*   - shapes.c  (Bill Sherman's simple OpenGL shapes)                 */
/*                                                                     */
/* Copyright 2007, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>	/* needed for printf(), etc. */
#include <stdlib.h>	/* needed for exit() */
#include <signal.h>	/* needed for signal() and its arguments */
#include <unistd.h>	/* not sure why/whether this is needed */
#include <math.h>	/* needed for fabs() */

#include <GL/gl.h>	/* needed for all OpenGL functions & values */

#include "freevr.h"	/* needed for FreeVR functions & values */
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */


/* Define a structure for information about each individual object. */
typedef struct {
		float				x, y, z;	/* location of the object   */
		float				size;		/* size of the object       */
		GLubyte				color[3];	/* color of the object      */
		enum { cube,pyramid,ud_pyramid }shape;		/* shape of the object      */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		object_count;			/* number of objects in world */
		int		selected_object;		/* currently selected object  */
		ObjectType	object[3];			/* array of objects in world  */
		vrLock		lock;				/* to avoid bad multiprocess accesses */
	} WorldDataType;


/*************************************************************************************/
/* These function declarations are very important -- they will be used as arguments. */
void	exit_application(int signal);		/* cleanly close down operations   */

void	init_gfx(void);				/* initialize the graphics         */
void	draw_world(WorldDataType *wd, vrRenderInfo *rendinfo); /* render the world */
void	init_world(WorldDataType *wd);		/* initialize the world parameters */
void	update_world(WorldDataType *wd);	/* perform the world simulation    */


/***********************************************************************/
main(int argc, char* argv[])
{
	WorldDataType	*wd;

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, SIG_IGN);
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */
	vrStart();
	signal(SIGINT, exit_application);

	/********************/
	/* create the world */
	wd = (WorldDataType *)vrShmemAlloc0(sizeof(WorldDataType));
	wd->lock = vrLockCreate();
	init_world(wd);

	/***************************************/
	/* set up the FreeVR library functions */
	vrFunctionSetCallback(VRFUNC_DISPLAY_INIT, vrCallbackCreate(init_gfx, 0));
	vrFunctionSetCallback(VRFUNC_DISPLAY, vrCallbackCreate(draw_world, 1, wd));

	/********************************************************/
	/* define the application name, authors, controls, etc. */
	vrSystemSetName("example 7 -- travel (C++ Version)");
	vrSystemSetAuthors("Bill Sherman");
	vrSystemSetExtraInfo("An example to demonstrate traveling relative to the virtual world");
	vrSystemSetStatusDescription("Application running fine");
	vrInputSet2switchDescription(0, "Terminate the application");
	vrInputSet2switchDescription(1, "Move the selected object");
	vrInputSet2switchDescription(3, "Reset the travel matrix (take us home)");
	vrInputSetValuatorDescription(0, "Rotate us in the world about the Y-axis");
	vrInputSetValuatorDescription(1, "Move us though the world in the direction we point");
	vrInputSet6sensorDescription(WAND_SENSOR, "Move a short cyan pointer to manipulate objects & choose travel dir");

	/***************************/
	/*** do world simulation ***/
	/***************************/
	vrUserTravelReset(VR_ALLUSERS);
	/* run until terminate key is pressed (Escape by default) */
	while(!vrGet2switchValue(0)) {
		vrFrame();
		update_world(wd);
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int signal)
{
	vrExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/


/**************************************************************/
/* init_world(): create the initial conditions of the dynamic */
/*   simulation.                                              */
/**************************************************************/
void init_world(WorldDataType *wd)
{
	vrLockWriteSet(wd->lock);

	/* object 1 - a blue cube */
	wd->object[0].shape = ObjectType::cube;
	wd->object[0].color[0] = 100;
	wd->object[0].color[1] = 100;
	wd->object[0].color[2] = 255;
	wd->object[0].x =  0.0;
	wd->object[0].y =  5.0;
	wd->object[0].z = -4.0;

	/* object 2 - a red pyramid */
	wd->object[1].shape = ObjectType::pyramid;
	wd->object[1].color[0] = 255;
	wd->object[1].color[1] = 100;
	wd->object[1].color[2] = 100;
	wd->object[1].x = -4.0;
	wd->object[1].y =  5.0;
	wd->object[1].z = -4.0;

	/* object 3 - an upside-down green pyramid */
	wd->object[2].shape = ObjectType::ud_pyramid;
	wd->object[2].color[0] = 100;
	wd->object[2].color[1] = 255;
	wd->object[2].color[2] = 100;
	wd->object[2].x =  4.0;
	wd->object[2].y =  5.0;
	wd->object[2].z = -4.0;

	wd->object_count = 3;
	wd->selected_object = -1;

	vrLockWriteRelease(wd->lock);
}


/*********************************************************************/
/* update_world(): In this version, we handle the user interactions. */
/*   Both selecting, and moving the selected object.                 */
/*********************************************************************/
void update_world(WorldDataType *wd)
{
#define JS_EPSILON	0.125				/* the dead-zone before joystick movement is recognized */
#define MOVE_FACTOR	2.5				/* a scale factor to tune travel movement               */
#define TURN_FACTOR	25.0				/* a scale factor to tune travel movement               */
#define SUB_JS_EPSILON(a)	((a)-copysign(JS_EPSILON,(a)))	/* a macro to remove the joystick epsilon range */
static	vrTime		last_time = -1.0;		/* initially set to an invalid value as a flag */
	vrTime		sim_time = vrCurrentSimTime();	/* the current time of the simulation  */
static	vrPoint		last_wand_vw_locpnt;		/* VW location of the wand in the last update frame */
	vrPoint		wand_vw_locpnt;			/* current location of the wand (in VW coords) */
#define	last_wand_vw_loc last_wand_vw_locpnt.v
#define	wand_vw_location wand_vw_locpnt.v
	vrVector	wand_rw_pointvec;		/* current vector pointing out of the wand (in RW coords) */
	double		delta_time;			/* time since last update (in seconds) */
	double		delta_move;			/* distance an object should be moved */
	double		joy_x, joy_y;			/* Joystick values */
	int		selected_obj;			/* the currently selected object  */
	int		touch_obj = -1;			/* the currently touched object   */
	int		objnum;				/* for looping throuh the objects */
	ObjectType	*object;			/* pointer to object to be moved  */

	/*****************************************************/
	/** Determine delta time from last simulation frame **/
	/*****************************************************/
	if (last_time == -1.0)
		delta_time = 0.0;
	else	delta_time = sim_time - last_time;
	last_time = sim_time;				/* now that delta_time has been calculated, we won't use last_time until next time */

	/* skip the update if the delta isn't big enough */
	if (delta_time <= 0.0)	/* can also choose a non-zero epsilon */
		return;

	/**********************************/
	/** Handle Manipulation via wand **/
	/**********************************/

	/**********************************************************/
	/* check for contact between the wand and all the objects */
	/* Note, this is very course locking.  We could also lock just */
	/*   around the lines that actually access the world database. */
	vrLockReadSet(wd->lock);

	vrPointGetVWFromUser6sensor(&wand_vw_locpnt, 0, WAND_SENSOR);

	for (objnum = 0; objnum < wd->object_count; objnum++) {
		if ( (wand_vw_location[VR_X] < wd->object[objnum].x + 1.0)
		  && (wand_vw_location[VR_X] > wd->object[objnum].x - 1.0)
		  && (wand_vw_location[VR_Y] < wd->object[objnum].y + 1.0)
		  && (wand_vw_location[VR_Y] > wd->object[objnum].y - 1.0)
		  && (wand_vw_location[VR_Z] < wd->object[objnum].z + 1.0)
		  && (wand_vw_location[VR_Z] > wd->object[objnum].z - 1.0) ) {
			touch_obj = objnum;
		}
	}

	selected_obj = wd->selected_object;

	vrLockReadRelease(wd->lock);

	if (selected_obj != touch_obj) {
		/* NOTE: a local variable (touch_obj) was used because it */
		/*   decreases the time the lock must be held.            */
		vrLockWriteSet(wd->lock);
		wd->selected_object = touch_obj;
		vrLockWriteRelease(wd->lock);

		selected_obj = touch_obj;
	}

	/* when left wand button is pressed, the selected object is moved */
	if (vrGet2switchValue(1) && (selected_obj != -1)) {
		vrLockWriteSet(wd->lock);
		object = &(wd->object[selected_obj]);
		object->x += (wand_vw_location[VR_X] - last_wand_vw_loc[VR_X]);
		object->y += (wand_vw_location[VR_Y] - last_wand_vw_loc[VR_Y]);
		object->z += (wand_vw_location[VR_Z] - last_wand_vw_loc[VR_Z]);
		vrLockWriteRelease(wd->lock);
	}

	/******************************************************************/
	/* store the current wand location for next time through the loop */
	last_wand_vw_loc[VR_X] = wand_vw_location[VR_X];
	last_wand_vw_loc[VR_Y] = wand_vw_location[VR_Y];
	last_wand_vw_loc[VR_Z] = wand_vw_location[VR_Z];


   /**** begin new code from example 6 ****/
	/****************************/
	/** Handle Travel via wand **/
	/****************************/

	/* pressing the right wand button resets us to the initial position */
	if (vrGet2switchValue(3)) {
		vrUserTravelReset(VR_ALLUSERS);
	}

	/* use wand joystick to fly through world */
	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	if (fabs(joy_x) > JS_EPSILON)
		vrUserTravelRotateId(VR_ALLUSERS, VR_Y, (delta_time * SUB_JS_EPSILON(joy_x) * -TURN_FACTOR));

	if (fabs(joy_y) > JS_EPSILON) {
		delta_move = delta_time * SUB_JS_EPSILON(joy_y) * MOVE_FACTOR;
		vrVectorGetRWFrom6sensorDir(&wand_rw_pointvec, WAND_SENSOR, VRDIR_FORE);
		vrUserTravelTranslate3d(VR_ALLUSERS,
			wand_rw_pointvec.v[VR_X] * delta_move,
			wand_rw_pointvec.v[VR_Y] * delta_move,
			wand_rw_pointvec.v[VR_Z] * delta_move);
	}
   /**** end new code ****/

	/*******************************************/
	/* allow other processes access to the CPU */
	vrSleep(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

#ifdef __cplusplus
extern "C" {
#endif

/* functions from shapes.c */
void	draw_floor();
void	draw_cube();
void	draw_pyramid();
void	draw_cube_outline();
void	draw_pyramid_outline();

#ifdef __cplusplus
}
#endif

/*********************************************************************/
/* init_gfx(): initialize general graphics properties (eg. lighting) */
/*********************************************************************/
void init_gfx(void)
{
	GLfloat	location_0[4] = { 0.0, 10.0, 10.0, 1.0 /* 0 = directional, 1 = location */},
		ambient_0[4] = { 0.2, 0.2, 0.2, 1.0 },
		diffuse_0[4] = { 0.8, 0.8, 0.8, 1.0 },
		specular_0[4] = { 0.2, 0.2, 0.2, 1.0 };

vrMsgPrintf("App: " BOLD_TEXT "****************** Entering init_gfx().  When am I?\n" NORM_TEXT);

	/********************************************************/
	/* set the polygon shading/lighting/material parameters */
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0, GL_POSITION, location_0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular_0);
	glEnable(GL_LIGHT0);

	/* These two lines cause the vertex colors to be used (otherwise all is grey) */
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_NORMALIZE);	/* this is required when calling glScalef() w/ lighting */

	/***************************/
	/* set the line parameters */
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(2.0);
}


/*******************************************************************/
/* draw_world(): Now will render how ever many objects are in the  */
/*   world data structure.  The color, type (and maybe rotation)   */
/*   are all now stored for each object, in addition to location.  */
/*   The currently selected object is highlighted by adding a      */
/*   wireframe outline to the representation.                      */
/*******************************************************************/
void draw_world(WorldDataType *wd, vrRenderInfo *rendinfo)
{
static	GLfloat		location_0[4] = { 0.0, 10.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
static	char		msg[128];			/* memory for text string messages */
	int		objnum;				/* object counter */

	/**************************************/
	/* clear the screen -- very important */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/*****************************************************************************/
	/* update the light's location -- must be done independently for each screen */
	glLightfv(GL_LIGHT0, GL_POSITION, location_0);

	/************************************/
	/* draw a short pointer at the wand */
	glPushMatrix();
		glColor3ub(100, 255, 255);	/* cyan */
		vrRenderTransform6sensor(rendinfo, WAND_SENSOR);
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glScalef(0.15, 0.15, 0.15);
		glTranslatef(0.0, 1.5, 0.0);
		draw_pyramid_outline();
	glPopMatrix();

	/*******************************************************/
	/* draw an object that remains fixed in physical space */
	glColor3ub(200, 100, 255);	/* purple */
	glPushMatrix();
		glTranslatef(-4.0, 1.0, -1.0);
		draw_cube();
	glPopMatrix();

	/******************************************************************/
	/* draw the rest of the objects in virtual (user navigated) space */
	glPushMatrix();		/* save the real-world coordinate system */
	vrRenderTransformUserTravel(rendinfo);

	draw_floor();

	/* Note again, this is very course locking. */
	vrLockReadSet(wd->lock);
	for (objnum = 0; objnum < wd->object_count; objnum++) {

		glColor3ubv(wd->object[objnum].color);
		glPushMatrix();
			glTranslatef(wd->object[objnum].x, wd->object[objnum].y, wd->object[objnum].z);
			switch (wd->object[objnum].shape) {
			case ObjectType::cube:
				if (objnum == wd->selected_object)
					draw_cube_outline();
				else	draw_cube();
				break;
			case ObjectType::pyramid:
				if (objnum == wd->selected_object)
					draw_pyramid_outline();
				else	draw_pyramid();
				break;
			case ObjectType::ud_pyramid:
				glScalef(1.0, -1.0, 1.0);
				if (objnum == wd->selected_object)
					draw_pyramid_outline();
				else	draw_pyramid();
				break;
			}
		glPopMatrix();
	}

	glPopMatrix();	/* return to real-world coordinates */
	/* NOTE: 2-D text is best rendered in a fixed location in the world */

	/*************************************************/
	/* draw text indicating which object is selected */
	if (wd->selected_object < 0) {
		sprintf(msg, "Object selected: none");
	} else {
		sprintf(msg, "Object selected: %d", wd->selected_object);
	}

	vrLockReadRelease(wd->lock);

	glDisable(GL_LIGHTING);		/* make the text bright regardless of lights */
	glColor3ub(255, 255, 255);	/* white */
	glPushMatrix();
		glTranslatef(-4.0, 6.0, -5.0);	/* on the front screen */

		glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
		vrRenderText(rendinfo, msg);
	glPopMatrix();
	glEnable(GL_LIGHTING);
}

