/***********************************************************************/
/* FreeVR OpenGL tutorial example 21: ex21_flashlight.c                */
/* Last tested with: FreeVR version 0.6a                               */
/* Last Modified: October 13, 2005                                     */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/* This example code demonstrates a VR application program that allows */
/*   the user to control lighting in the world.  The user has the      */
/*   ability to switch between multiple styles of OpenGL lighting.     */
/*                                                                     */
/*                                                                     */
/* The new features of this example include:                           */
/*   - the use of the vrRenderInfo structure in the VR_DISPLAY_INIT callback (though just for debugging output) */
/*   - ... */
/*                                                                     */
/*                                                                     */
/* Considerations:                                                     */
/*   - for good lighting across larger shapes, it is important to break*/
/*       them up into smaller polygons, to allow the OpenGL renderer   */
/*       to calculate the lighting as it applies to individual areas   */
/*       of an object.                                                 */
/*   - This application adds two new uses for button presses, while    */
/*       keeping the previous functionality.  Since not all systems    */
/*       will have more than 3 readily available buttons, two of the   */
/*       older operations have been moved to buttons 4 & 5, allowing   */
/*       the lighting control operations access to buttons 1 & 2.      */
/*                                                                     */
/* This application uses the files:                                    */
/*   - wrs_shapes.c  (Bill Sherman's simple OpenGL shapes)             */
/*                                                                     */
/* Copyright 2010, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>	/* needed for printf(), etc. */
#include <stdlib.h>	/* needed for exit() & getenv() */
#include <signal.h>	/* needed for signal() and its arguments */
#include <unistd.h>	/* not sure why/whether this is needed */
#include <math.h>	/* needed for fabs() */
#include <inttypes.h>	/* needed for "uint32_t" type */

#include <GL/gl.h>	/* needed for all OpenGL functions & values */

#include "freevr.h"	/* needed for FreeVR functions & values */
#define HEAD_SENSOR	0
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */

/* Define some values to adjust the compilation of this example */
#define	USE_SOCKETS
#define MAX_OBJECTS	40

#define NO_TEXT     0
#define X_2DTEXT    1
#define GLUT_2DTEXT 2
#define GLUT_3DTEXT 3
#define TEXT_STYLE X_2DTEXT

#if (TEXT_STYLE == GLUT_2DTEXT || TEXT_STYLE == GLUT_3DTEXT)
#  include <GL/glut.h>	/* needed for Glut functions -- primarily text rendering */
#endif


/* Define a structure for information about each individual object. */
typedef struct {
		float			x, y, z;		/* location of the object     */
		float			size;			/* size of the object         */
		GLubyte			color[3];		/* color of the object        */
		GLuint			texture_id;		/* texture ID code for object */
		char			*msg;			/* text of a text object      */
		enum {	cube,
			tex_cube,
			tex_plane,
			sphere,
			pyramid,
			ud_pyramid,
			billboard_c,
			billboard_s,
			text }		shape;			/* shape of the object        */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		quit_bit;			/* a flag to pass along the desire to terminate */
		int		number_of_objects;		/* number of objects in world */
		int		selected_object;		/* currently selected object  */
		int		show_wim;			/* flag indicating whether WIM should be rendered */
		double		cutplane[4];			/* A,B,C & D plane parameters for the cut-away cutting plane */
		int		light_in_hand;			/* choice of the type of light */
		int		scene_light_on;			/* flag indicating whether the global light is turned on */
		ObjectType	object[MAX_OBJECTS];		/* array of objects in world  */
		vrLock		lock;				/* to avoid bad multiprocess accesses */
	} WorldDataType;


/*************************************************************************************/
/* These function declarations are very important -- they will be used as arguments. */
void	exit_application(int signal);		/* cleanly close down operations     */

void	init_gfx();				/* initialize the graphics           */
void	draw_world(WorldDataType *wd, vrRenderInfo *rendinfo); /* render the world   */
void	init_world(WorldDataType *wd);		/* initialize the world parameters   */
void	update_world(WorldDataType *wd);	/* perform the world simulation      */

void	end_world();				/* close down the world operations   */
void	init_objects(WorldDataType *wd);	/* initalize the objects in the world*/


/***********************************************************************/
main(int argc, char* argv[])
{
	WorldDataType	*wd;

#ifdef FREEGLUT
#  if (TEXT_STYLE == GLUT_2DTEXT || TEXT_STYLE == GLUT_3DTEXT)
	/* NOTE: The FreeGlut version of Glut introduced the unnecessary */
	/*   requirement of calling glutInit(), even when the only use   */
	/*   of the Glut library is for rendering objects, and not using */
	/*   the windowing or input side of Glut.  Hopefully, one day,   */
	/*   either the FreeGlut project will eliminate this odd         */
	/*   requirement, or Linux distributions will shift to a         */
	/*   different Glut distribution.                                */
	glutInit(&argc, argv);
#  endif
#endif

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, SIG_IGN);
	vrShmemInit(1024*1024*8);		/* needed for multiple screens because of the texture memory */
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */
	vrStart();
	signal(SIGINT, exit_application);

	/********************/
	/* create the world */
	wd = (WorldDataType *)vrShmemAlloc0(sizeof(WorldDataType));
	wd->lock = vrLockCreate();
	init_world(wd);

	/***************************************/
	/* set up the FreeVR library functions */
	vrFunctionSetCallback(VRFUNC_DISPLAY_INIT, vrCallbackCreate(init_gfx, 0));
	vrFunctionSetCallback(VRFUNC_DISPLAY, vrCallbackCreate(draw_world, 1, wd));

	/********************************************************/
	/* define the application name, authors, controls, etc. */
	vrSystemSetName("example 21 -- interactive light example");
	vrSystemSetAuthors("Bill Sherman");
	vrSystemSetExtraInfo("An example of an interactive light in VR");
	vrSystemSetStatusDescription("Application running fine");
	vrInputSet2switchDescription(0, "Terminate the application");
	vrInputSet2switchDescription(4, "Move the selected object");
	vrInputSet2switchDescription(5, "Turn on the WIM display");
	vrInputSet2switchDescription(3, "Reset the travel matrix (take us home)");
	vrInputSet2switchDescription(1, "Toggle the scene light");
	vrInputSet2switchDescription(2, "Rotate through the handheld light options");
	vrInputSetValuatorDescription(0, "Rotate us in the world about the Y-axis");
	vrInputSetValuatorDescription(1, "Move us though the world in the direction we point");
	vrInputSet6sensorDescription(WAND_SENSOR, "Move a short cyan pointer to manipulate objects & choose travel dir");

	/***************************/
	/*** do world simulation ***/
	/***************************/
	vrUserTravelReset(VR_ALLUSERS);
	/* run until the user has indicated it's time not to */
	while(!wd->quit_bit) {
		vrFrame();
		update_world(wd);
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int signal)
{
	end_world();

	vrExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

#ifdef USE_SOCKETS /* { */
static int	listen_sock = -1;		/* socket to receive commands */
						/*   listen_sock is not in wd        */
						/*   because it needs to be          */
						/*   available to exit_application() */
						/*   which can have no arguments.    */
#endif /* } USE_SOCKETS */


/****************************************************************/
/* init_world(): create the initial conditions of the dynamic   */
/*   simulation.  If the "VR_PORT" environment variable is set, */
/*   then its value is used as the initial server socket port.  */
/*   Otherwise, the arbitrarily chosen port number 4000 is used.*/
/*   Since IRIX, and perhaps other Unicies have a "feature" in  */
/*   which a socket port isn't usable again until after 60      */
/*   seconds have passed since the connection was closed, if the*/
/*   initial socket isn't available, we try the next two ports  */
/*   in succession.  The selected port is reported to the tty.  */
/****************************************************************/
void init_world(WorldDataType *wd)
{
#ifdef USE_SOCKETS /* { */
	int	inet_port;
	int	inet_port_max;
	char	host_machine[128];
#endif /* } USE_SOCKETS */

	/*************************/
	/* set world state flags */
	vrLockWriteSet(wd->lock);
	wd->quit_bit = 0;
	wd->show_wim = 0;
	vrLockWriteRelease(wd->lock);

	/***************************/
	/* setup the world content */
	init_objects(wd);

#ifdef USE_SOCKETS /* { */
	/******************************/
	/* initialize the socket port */
	gethostname(host_machine, 128);

	/* loop through N possible internet ports to hopefully find one that works */
	if (getenv("VR_PORT") != NULL)
		inet_port = atoi(getenv("VR_PORT"));
	else	inet_port = 4000;
	for (inet_port_max = inet_port + 2; listen_sock < 0 && inet_port <= inet_port_max; inet_port++) {
		printf("Port %d -- ", inet_port);
		fflush(stdout);
		listen_sock = vrSocketCreateListen(&inet_port, 0);
	}
	inet_port--;
	if (listen_sock < 0)
		printf("WARNING: Unable to find an open socket port.\n");
	else	printf("\rConnect to '%s:%d' to send commands to the simulation\n", host_machine, inet_port);
#endif /* } USE_SOCKETS */
}


/***************************************************************/
/* init_objects(): place objects into the world data structure */
/*   at some initial positions.                                */
/***************************************************************/
void init_objects(WorldDataType *wd)
{
	int	objnum = 0;

	vrLockWriteSet(wd->lock);

#if 0
	/* object 1 - a blue cube */
	wd->object[objnum].shape = cube;
	wd->object[objnum].size = 1.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 255;
	wd->object[objnum].x =  0.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;
#endif

	/* object 2 - a red pyramid */
	wd->object[objnum].shape = pyramid;
	wd->object[objnum].size = 1.0;
	wd->object[objnum].color[0] = 255;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x = -10.0;
	wd->object[objnum].y =   1.5;
	wd->object[objnum].z = -10.0;
	objnum++;

#if 0
	/* object 3 - an upside-down green pyramid */
	wd->object[objnum].shape = ud_pyramid;
	wd->object[objnum].size = 1.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x =  4.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;
#endif

	/* object 4 - some green text */
	wd->object[objnum].shape = text;
	wd->object[objnum].size = 1.0;
	wd->object[objnum].color[0] = 255;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x = -10.0;
	wd->object[objnum].y =   2.5;
	wd->object[objnum].z = -10.0;
	wd->object[objnum].msg = vrShmemStrDup("destination");
	objnum++;

	/* Make a simple maze out of cubes and walls */
	wd->number_of_objects = 4;
	wd->object[objnum].shape = tex_plane;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x = -10.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z =  -7.0;
	objnum++;

	wd->object[objnum].shape = tex_cube;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x = -4.0;
	wd->object[objnum].y =  3.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	wd->object[objnum].shape = tex_cube;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x = -16.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z = -10.0;
	objnum++;

	wd->object[objnum].shape = tex_cube;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x =   2.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z = -10.0;
	objnum++;

	wd->object[objnum].shape = tex_cube;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x = -10.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z = -16.0;
	objnum++;

	wd->object[objnum].shape = tex_plane;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x =  -4.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z = -19.0;
	objnum++;

	wd->object[objnum].shape = tex_plane;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x =   2.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z = -19.0;
	objnum++;

	wd->object[objnum].shape = tex_plane;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x =   8.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z = -19.0;
	objnum++;

	wd->object[objnum].shape = tex_cube;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x =  14.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z = -16.0;
	objnum++;

	wd->object[objnum].shape = tex_cube;
	wd->object[objnum].size = 3.0;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].texture_id = 1;
	wd->object[objnum].x =  14.0;
	wd->object[objnum].y =   3.0;
	wd->object[objnum].z = -10.0;
	objnum++;

	wd->number_of_objects = objnum;
	wd->selected_object = -1;

	/* sanity test */
	if (objnum > MAX_OBJECTS) {
		printf("DANGER: Number of objects (%d), exceeds maximum allowable (%d).  You may experience issues.\n", objnum, MAX_OBJECTS);
	}

	wd->light_in_hand = 0;
	wd->scene_light_on = 1;

	wd->cutplane[VR_X] =  0.0;
	wd->cutplane[VR_Y] =  0.0;
	wd->cutplane[VR_Z] = -1.0;
	wd->cutplane[VR_W] =  0.0;

	vrLockWriteRelease(wd->lock);
}


/*******************************************************/
/* end_world(): clean up anything opened by simulation */
/*******************************************************/
void end_world()
{
#ifdef USE_SOCKETS /* { */
	vrSocketClose(listen_sock);
#endif /* } USE_SOCKETS */
}


/*********************************************************************/
/* update_world(): In this version, we handle the user interactions. */
/*   Both selecting, and moving the selected object.                 */
/*********************************************************************/
void update_world(WorldDataType *wd)
{
#define JS_EPSILON	0.125				/* the dead-zone before joystick movement is recognized */
#define MOVE_FACTOR	2.5				/* a scale factor to tune travel movement               */
#define TURN_FACTOR	25.0				/* a scale factor to tune travel movement               */
#define SUB_JS_EPSILON(a)	((a)-copysign(JS_EPSILON,(a)))	/* a macro to remove the joystick epsilon range */
#ifdef USE_SOCKETS /* { */
static	char	*help_msg = "Socket commands:\n"
				"\thelp -- this message\n"
				"\tquit -- quit the application\n"
				"\tclose -- close the connection\n"
				"\treset -- reset the application\n"
				"\tmove cube up -- translate the cube up 2 feet\n"
				"\tmove cube down -- translate the cube down 2 feet\n"
				"\twhere -- report the user's position in the world\n"
				"\twhere am i -- report the user's position in the world\n"
				"\tshow ActiveWalls -- report some CAVE config info\n"
		"";
static	int	cmd_socket = -1;			/* socket to get commands from */
static	char	buf[128];				/* socket read buffer          */
static	char	msg[128];				/* socket send buffer          */
	int	result;					/* result of reading a socket  */
#endif /* } USE_SOCKETS */

static	vrTime		last_time = -1.0;		/* initially set to an invalid value as a flag */
	vrTime		sim_time = vrCurrentSimTime();	/* the current time of the simulation  */
static	vrPoint		last_wand_vw_locpnt;		/* VW location of the wand in the last update frame */
	vrPoint		wand_rw_locpnt;			/* current location of the wand (in RW coords) */
	vrPoint		wand_vw_locpnt;			/* current location of the wand (in VW coords) */
#define	last_wand_vw_loc last_wand_vw_locpnt.v
#define	wand_vw_location wand_vw_locpnt.v
	vrVector	wand_rw_pointvec;		/* current vector pointing out of the wand (in RW coords) */
	vrVector	wand_vw_pointvec;		/* current vector pointing out of the wand (in VW coords) */
	double		delta_time;			/* time since last update (in seconds) */
	double		delta_move;			/* distance an object should be moved */
	double		plane_distance;			/* distance from the wand to the origin [? verify]*/
	double		joy_x, joy_y;			/* Joystick values */
	int		selected_object_num;		/* the object selected in the database */
	int		touch_obj = -1;			/* the currently touched object   */
	int		objnum;				/* for looping throuh the objects */
	ObjectType	*object;			/* pointer to object to be compared/moved  */
	float		size;				/* size of object -- used for isect comparison */

	/*****************************************************/
	/** Determine delta time from last simulation frame **/
	/*****************************************************/
	if (last_time == -1.0)
		delta_time = 0.0;
	else	delta_time = sim_time - last_time;
	last_time = sim_time;			/* now that delta_time has been calculated, we won't use last_time until next time */

	/* skip the update if the delta isn't big enough */
	if (delta_time <= 0.0)			/* can also choose a non-zero epsilon */
		return;

	/**********************************/
	/** Handle Manipulation via wand **/
	/**********************************/

	/**********************************************************/
	/* check for contact between the wand and all the objects */
	vrLockReadSet(wd->lock);

	vrPointGetRWFrom6sensor(&wand_rw_locpnt, WAND_SENSOR);
	vrVectorGetRWFrom6sensorDir(&wand_rw_pointvec, WAND_SENSOR, VRDIR_FORE);
	vrPointGetVWFromUser6sensor(&wand_vw_locpnt, 0, WAND_SENSOR);
	vrVectorGetVWFromUser6sensorDir(&wand_vw_pointvec, 0, WAND_SENSOR, VRDIR_FORE);

	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {
		object = &(wd->object[objnum]);
		size = object->size;

		if ( (wand_vw_location[VR_X] < object->x + size)
		  && (wand_vw_location[VR_X] > object->x - size)
		  && (wand_vw_location[VR_Y] < object->y + size)
		  && (wand_vw_location[VR_Y] > object->y - size)
		  && (wand_vw_location[VR_Z] < object->z + size)
		  && (wand_vw_location[VR_Z] > object->z - size) ) {
			touch_obj = objnum;
		}
	}

	selected_object_num = wd->selected_object;

	vrLockReadRelease(wd->lock);

	if (selected_object_num != touch_obj) {
		/* NOTE: a local variable (touch_obj) is used because it */
		/*   decreases the time the lock must be held.           */
		vrLockWriteSet(wd->lock);
		wd->selected_object = touch_obj;
		vrLockWriteRelease(wd->lock);
	}

	/* when #4 wand button is pressed, the selected object is moved */
	if (vrGet2switchValue(4) && (touch_obj != -1)) {
		vrLockWriteSet(wd->lock);
		object = &(wd->object[touch_obj]);
		object->x += (wand_vw_location[VR_X] - last_wand_vw_loc[VR_X]);
		object->y += (wand_vw_location[VR_Y] - last_wand_vw_loc[VR_Y]);
		object->z += (wand_vw_location[VR_Z] - last_wand_vw_loc[VR_Z]);
		vrLockWriteRelease(wd->lock);
	}

	/******************************************************************/
	/* store the current wand location for next time through the loop */
	last_wand_vw_loc[VR_X] = wand_vw_location[VR_X];
	last_wand_vw_loc[VR_Y] = wand_vw_location[VR_Y];
	last_wand_vw_loc[VR_Z] = wand_vw_location[VR_Z];

	/*****************************/
	/* update the clipping plane */
	plane_distance = -vrVectorDotProduct(&wand_rw_pointvec, (vrVector*)&wand_rw_locpnt);
	vrLockWriteSet(wd->lock);
	wd->cutplane[VR_X] = wand_rw_pointvec.v[VR_X];
	wd->cutplane[VR_Y] = wand_rw_pointvec.v[VR_Y];
	wd->cutplane[VR_Z] = wand_rw_pointvec.v[VR_Z];
	wd->cutplane[VR_W] = plane_distance;

	/*********************************/
	/* toggle the world-in-miniature */
	/* NOTE: write-lock is still set from clipping plane parameters (above) */
	wd->show_wim = vrGet2switchValue(5);
	vrLockWriteRelease(wd->lock);

	/****************************/
	/** Handle Travel via wand **/
	/****************************/

	/* use wand joystick to fly through world */
	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	if (fabs(joy_x) > JS_EPSILON)
		vrUserTravelRotateId(VR_ALLUSERS, VR_Y, (delta_time * SUB_JS_EPSILON(joy_x) * -TURN_FACTOR));

	if (fabs(joy_y) > JS_EPSILON) {
		delta_move = delta_time * SUB_JS_EPSILON(joy_y) * MOVE_FACTOR;
		vrUserTravelTranslate3d(VR_ALLUSERS,
			wand_rw_pointvec.v[VR_X] * delta_move,
			wand_rw_pointvec.v[VR_Y] * delta_move,
			wand_rw_pointvec.v[VR_Z] * delta_move);
	}


	/********************************/
	/* Handle the lighting commands */
	/********************************/

	/* NOTE: for the moment at least, I'm using an extra two buttons that are available on the IS-900 */

	/* toggle the scene light */
	if (vrGet2switchDelta(1) == 1) {
		wd->scene_light_on ^= 1;
		printf("Light is now %s\n", (wd->scene_light_on ? "on" : "off"));
		sprintf(msg, "Scene lighting is: %s, with light %d\n", (wd->scene_light_on ? "on" : "off"), wd->light_in_hand);
		vrSystemSetStatusDescription(msg);
	}

	if (vrGet2switchDelta(2) == 1) {
		wd->light_in_hand++;
		wd->light_in_hand %= 4;
		printf("Now using light %d\n", wd->light_in_hand);
		sprintf(msg, "Scene lighting is: %s, with light %d\n", (wd->scene_light_on ? "on" : "off"), wd->light_in_hand);
		vrSystemSetStatusDescription(msg);
	}

	/* pressing the right wand button resets us to the initial position */
	if (vrGet2switchValue(3)) {
		vrUserTravelReset(VR_ALLUSERS);
	}


#ifdef USE_SOCKETS /* { */
	/*******************************/
	/* Handle Commands from socket */

	/* if no connection (and a listen socket is open), see if someone is calling */
	if (cmd_socket < 0 && listen_sock >= 0) {
		/* cmd_socket will be -1 when no one is connected */
		cmd_socket = vrSocketAnswer(listen_sock);
		if (cmd_socket > 0) {
			printf("New client connected.\n");
			vrSocketSendMsg(cmd_socket, "Your wish is my command >");
		}
	}

	/* if a connection, see if client sent a command.  If so, act on it */
	if (cmd_socket > 0) {
		result = vrSocketReadMsgClean(cmd_socket, buf, 128);
		switch (result) {
		case -1:
			break;
		case  0:
			printf("Client stopped communicating.\n");
			close(cmd_socket);
			cmd_socket = -1;
			break;
		default:
			printf("socket cmd = '%s'\n", buf);
			/* parse command here -- very simple-minded parsing */
			if (strcmp(buf, "quit") == 0) {
				/* set the world for termination */
				wd->quit_bit = 1;
			} else
			if (strcmp(buf, "help") == 0 || strcmp(buf, "?") == 0) {
				/* give usage information */
				vrSocketSendMsg(cmd_socket, help_msg);
			} else
			if (strcmp(buf, "close") == 0) {
				/* close socket connection */
				printf("Client ending communication.\n");
				close(cmd_socket);
				cmd_socket = -1;
			} else
			if (strcmp(buf, "reset") == 0) {
				/* make the world as it was in the beginning */
				init_objects(wd);
				vrUserTravelReset(VR_ALLUSERS);
			} else
			if (strcmp(buf, "move cube up") == 0) {
				/* move the blue cube up 2.0 units */
				wd->object[0].y += 2.0;
			} else
			if (strcmp(buf, "move cube down") == 0) {
				/* move the blue cube down 2.0 units */
				wd->object[0].y -= 2.0;
			} else
			if (strcmp(buf, "where") == 0 || strcmp(buf, "where am i") == 0) {
				/* report the user's location */
				vrMatrix	*headmat;
				vrPoint		headloc;

				headmat = vrMatrixCreate();
				vrMatrixGetRWFromUserHead(headmat, 0);
				vrPointGetVWFromUserMatrix(&headloc, 0, headmat);
				sprintf(msg, "You are at (%f %f %f).\n", headloc.v[VR_X], headloc.v[VR_Y], headloc.v[VR_Z]);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			if (strcmp(buf, "show ActiveWalls") == 0) {
				/* report the type of graphics (VR) display */
				sprintf(msg, "Application has %d active walls.\n", vrContext->config->num_windows);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			/* default: */ {
				vrSocketSendMsg(cmd_socket, "Unknown command.\n");
			}

			/* prompt for another command (if socket still open) */
			if (cmd_socket > 0) {
				vrSocketSendMsg(cmd_socket, "Your wish is my command >");
			}
		}
	}
#endif /* } USE_SOCKETS */

	/*********************************/
	/* Handle Commands from keyboard */
	if (vrGet2switchValue(0)) {
		wd->quit_bit = 1;
	}

	/****************************************************/
	/* yield to allow other processes access to the CPU */
	vrSleep(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

/* functions from wrs_shapes.c */
void	draw_floor();
void	draw_cube();
void	draw_sphere();
void	draw_pyramid();
void	draw_plane();
void	draw_cube_outline();
void	draw_sphere_outline();
void	draw_pyramid_outline();
void	draw_plane_outline();

void	*make_texture_checker_pattern(int xsize, int ysize, int trans);
void	set_texture_image(void *imagedata);
void	*readPPM(char *filename);


/*********************************************************************/
/* init_gfx(): initialize general graphics properties (eg. lighting) */
/*********************************************************************/
void init_gfx()
{
	GLfloat	location_0[4] = { 0.0, 10.0, 10.0, 1.0 /* 0 = directional, 1 = location */},
		ambient_0[4] = { 0.2, 0.2, 0.2, 1.0 },
		diffuse_0[4] = { 0.8, 0.8, 0.8, 1.0 },
		specular_0[4] = { 0.2, 0.2, 0.2, 1.0 },
		direction[4] = { 0.0, 0.0, 1.0, 0.0 };
	void	*imagedata;


	/****************************************/
	/* set the render initialization values */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);

	/********************************************************/
	/* set the polygon shading/lighting/material parameters */
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0, GL_POSITION, location_0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular_0);
	glEnable(GL_LIGHT0);

	/* these are the lights that move */
	glLightfv(GL_LIGHT1, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT1, GL_SPECULAR, specular_0);
	glLightfv(GL_LIGHT1, GL_POSITION, location_0);

	glLightfv(GL_LIGHT2, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT2, GL_SPECULAR, specular_0);
	glLightfv(GL_LIGHT2, GL_POSITION, location_0);

	glLightfv(GL_LIGHT3, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT3, GL_SPECULAR, specular_0);
	glLightfv(GL_LIGHT3, GL_POSITION, location_0);
	glLightfv(GL_LIGHT3, GL_SPOT_DIRECTION, direction);
	glLightf(GL_LIGHT3, GL_SPOT_EXPONENT, 0);
	glLightf(GL_LIGHT3, GL_SPOT_CUTOFF, 50);

	/* These two lines cause the vertex colors to be used (otherwise all is grey) */
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_NORMALIZE);			/* this is required when calling glScalef() w/ lighting */

	/***************************/
	/* set the line parameters */
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(2.0);

	/*********************************************************/
	/* set the alpha parameters for see-through texture maps */ 
	glEnable(GL_ALPHA_TEST);		/* when using transparencies */
	glAlphaFunc(GL_GREATER, 0.5);

	/**********************************/
	/* create and set the texture map */
	imagedata = make_texture_checker_pattern(4, 4, 0);

	glBindTexture(GL_TEXTURE_2D, 1);
	set_texture_image(imagedata);

	/* No need to set the clipping plane, because we're not rendering yet anyway. */
}


/**********************************************************************/
/* render_virtualworld(): Render just the objects that are considered */
/*   a part of the virtual world.  Ie. excluding user interface stuff.*/
/**********************************************************************/
void render_virtualworld(WorldDataType *wd, vrRenderInfo *rendinfo)
{
	int		objnum;				/* object counter */
	ObjectType	*object;			/* pointer to current object */
	double		size;				/* size of current object */
	vrEuler		angles;				/* used for billboarding */

	/*****************************************/
	/* draw the objects in the virtual world */
	glPushMatrix();
		glScalef(5.0, 5.0, 5.0);
		draw_floor();
	glPopMatrix();

	vrLockReadSet(wd->lock);		/* lock while reading info about objects */
	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {
		object = &(wd->object[objnum]);

		glColor3ubv(object->color);
		glPushMatrix();
			size = object->size;

			glTranslatef(object->x, object->y, object->z);
			glScalef(size, size, size);

			switch (object->shape) {
			case cube:
				if (objnum == wd->selected_object)
					draw_cube_outline();
				else	draw_cube();
				break;
			case tex_cube:
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, object->texture_id);
				if (objnum == wd->selected_object)
					draw_cube_outline();
				else	draw_cube();
				glDisable(GL_TEXTURE_2D);
				break;
			case tex_plane:
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, object->texture_id);
				if (objnum == wd->selected_object)
					draw_plane();
				else	draw_plane_outline();
				glDisable(GL_TEXTURE_2D);
				break;
			case sphere:
				if (objnum == wd->selected_object)
					draw_sphere_outline();
				else	draw_sphere();
				break;
			case pyramid:
				if (objnum == wd->selected_object)
					draw_pyramid_outline();
				else	draw_pyramid();
				break;
			case ud_pyramid:
				glScalef(1.0, -1.0, 1.0);
				if (objnum == wd->selected_object)
					draw_pyramid_outline();
				else	draw_pyramid();
				break;
			case billboard_c:
				vrRenderGetBillboardAngles3d(rendinfo, &angles, object->x, object->y, object->z);
				glRotatef(angles.r[VR_AZIM], 0.0, 1.0, 0.0);
				glEnable(GL_TEXTURE_2D);
				if (objnum == wd->selected_object)
					draw_plane();
				else	draw_plane_outline();
				glDisable(GL_TEXTURE_2D);
				break;
			case billboard_s:
				vrRenderGetBillboardAngles3d(rendinfo, &angles, object->x, object->y, object->z);
				glRotatef(angles.r[VR_AZIM], 0.0, 1.0, 0.0);
				glRotatef(angles.r[VR_ELEV], 1.0, 0.0, 0.0);
				glEnable(GL_TEXTURE_2D);
				if (objnum == wd->selected_object)
					draw_plane();
				else	draw_plane_outline();
				glDisable(GL_TEXTURE_2D);
				break;
			case text:
				glDisable(GL_LIGHTING);
				glRasterPos3f(0.0, 0.0, 0.0);
#if TEXT_STYLE == X_2DTEXT /* { */
				/* Render text as a 2D bitmap using X-windows */
				/* NOTE: if the beginning of the string is clipped, */
				/*   none of the characters will appear.            */
				vrRenderText(rendinfo, object->msg);
#elif TEXT_STYLE == GLUT_2DTEXT
				/* Render text as a 2D bitmap using the GLUT library */
				/* NOTE: if the beginning of the string is clipped, */
				/*   none of the characters will appear.            */
				{
				char	*chr;
				for (chr = object->msg; *chr; chr++)
					glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *chr);
				}
#elif TEXT_STYLE == GLUT_3DTEXT
				/* Render text as a flat, but 3D stroke font using the GLUT library */
				glScalef(0.01, 0.01, 0.01);
				{
				char	*chr;
				for (chr = object->msg; *chr; chr++)
					glutStrokeCharacter(GLUT_STROKE_ROMAN, *chr);
				}
#endif /* } TEXT_STYLE */
				glEnable(GL_LIGHTING);
				break;
			}
		glPopMatrix();
	}
	vrLockReadRelease(wd->lock);
}

/***************************************************************************/
/* draw_world(): Render all the objects fixed in real world positoins,     */
/*   and objects located in the virtual world (via render_virtualworld()). */
/***************************************************************************/
void draw_world(WorldDataType *wd, vrRenderInfo *rendinfo)
{
static	GLfloat		location_0[4] = { 0.0, 10.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
	int		show_wim;
	int		scene_light;
	int		hand_light;

	/**************************************/
	/* clear the screen -- very important */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	/********************************************************************************/
	/* update the light's location -- must be done independently for each rendering */
	if (scene_light) {
		GLfloat	location_0[4] = { 0.0, 10.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
		glEnable(GL_LIGHT0);
		glLightfv(GL_LIGHT0, GL_POSITION, location_0);
	} else {
		glDisable(GL_LIGHT0);
	}

	/**************************************************************/
	/* draw a short pointer, miniature world or light at the wand */
	vrLockReadSet(wd->lock);
	show_wim = wd->show_wim;
	scene_light = wd->scene_light_on;
	hand_light = wd->light_in_hand;
	vrLockReadRelease(wd->lock);

	glPushMatrix();
		glColor3ub(100, 255, 255);	/* cyan */
		vrRenderTransform6sensor(rendinfo, WAND_SENSOR);
		if (show_wim) {
			/* the world-in-miniature replaces the wand avatar */
			glScalef(0.025, 0.025, 0.025);
			render_virtualworld(wd, rendinfo);

			/* draw a sphere at the user's location */
			glColor3ub(100, 255, 255);	/* cyan */

			/* factor OUT the head relative to the world */
			vrRenderTransformUserTravelInv(rendinfo);
			vrRenderTransform6sensor(rendinfo, HEAD_SENSOR);
			draw_sphere();
		}

		if (hand_light != 0) {
			GLfloat	position_x[4] = { 0.0, 0.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
			GLfloat	direction_x[4] = { 0.0, 0.0,  1.0, 0.0 /* 0 = directional, 1 = location */};
			GLfloat	spot_direction_x[4] = { 0.0, 0.0, -1.0, 0.0 /* 0 = directional, 1 = location */};

			/* draw a light that follows the wand */
			switch (hand_light) {
			case 1:
				glLightfv(hand_light+GL_LIGHT0, GL_POSITION, position_x);
				break;
			case 2:
				glLightfv(hand_light+GL_LIGHT0, GL_POSITION, direction_x);
				break;
			case 3:
				glLightfv(hand_light+GL_LIGHT0, GL_POSITION, position_x);
				glLightfv(hand_light+GL_LIGHT0, GL_SPOT_DIRECTION, spot_direction_x);
				break;
			}
			glEnable(hand_light+GL_LIGHT0);
		}

		if (!hand_light && !show_wim) {
			/* draw the standard wand avatar */
			glRotatef(-90.0, 1.0, 0.0, 0.0);
			glScalef(0.15, 0.15, 0.15);
			glTranslatef(0.0, 1.5, 0.0);
			draw_pyramid_outline();
		}
	glPopMatrix();

	/*******************************************************/
	/* draw an object that remains fixed in physical space */
	glColor3ub(200, 100, 255);		/* purple */
	glPushMatrix();
		glTranslatef(-4.0, 1.0, -1.0);
		draw_cube();
	glPopMatrix();

#if 0 /* for this example we disable the clipping plane { */
	/*****************************************************/
	/* draw a plane object, and then clip half the world */
	glPushMatrix();
		glColor3ub(255, 255, 255);	/* white */
		vrRenderTransform6sensor(rendinfo, WAND_SENSOR);
		glTranslatef(0.0, 0.0, 0.0);
		draw_plane_outline_only();
	glPopMatrix();
	glPushMatrix();
		vrLockReadSet(wd->lock);
		glClipPlane(GL_CLIP_PLANE0, wd->cutplane);
		vrLockReadRelease(wd->lock);
		glEnable(GL_CLIP_PLANE0);
	glPopMatrix();
#endif /* } */

	/******************************************/
	/* Now render the full-size virtual world */
	vrRenderTransformUserTravel(rendinfo);	/* change to virtual-world coordinates */
	render_virtualworld(wd, rendinfo);

#if 0 /* for this example we never enable the clipping plane */
	glDisable(GL_CLIP_PLANE0);
#endif

	if (hand_light != 0) {
		glDisable(hand_light+GL_LIGHT0);
	}
}

