/***********************************************************************/
/* FreeVR OpenGL tutorial example 7: ex7_pointer.c                     */
/* Last tested with: FreeVR version 0.6a                               */
/* Last Modified: January 29, 2007                                     */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/* This example code demonstrates a VR application program that        */
/*   renders a pointer emanating from the wand.  Everything else is    */
/*   the same.                                                         */
/*                                                                     */
/*                                                                     */
/* The new features of this example include:                           */
/*   - FreeVR Library functions, variables, and macros:                */
/*      * vrRenderTransform6sensor(rendinfo, )                         */
/*   - drawing a pointer emanating from the wand                       */
/*   - uses a few more OpenGL calls                                    */
/*   - more advanced structure to the world-database                   */
/*                                                                     */
/* Considerations:                                                     */
/*                                                                     */
/* This application uses the files:                                    */
/*   - shapes.c  (Bill Sherman's simple OpenGL shapes)                 */
/*                                                                     */
/* Copyright 2010, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>	/* needed for printf(), etc. */
#include <stdlib.h>	/* needed for exit() */
#include <signal.h>	/* needed for signal() and its arguments */
#include <unistd.h>	/* not sure why/whether this is needed */
#include <math.h>	/* needed for sin(), cos(), fabs() & copysign() */

#include <GL/gl.h>	/* needed for all OpenGL functions & values */

#include "freevr.h"	/* needed for FreeVR functions & values */
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */


/* Define a structure for information about each individual object. */
typedef struct {
		float				x, y, z;	/* location of the object   */
		GLubyte				color[3];	/* color of the object      */
		enum { cube,pyramid,ud_pyramid }shape;		/* shape of the object      */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		number_of_objects;		/* number of objects in world */
		ObjectType	object[3];			/* array of objects in world  */
	} WorldDataType;


/*************************************************************************************/
/* These function declarations are very important -- they will be used as arguments. */
void	exit_application(int signal);		/* cleanly close down operations     */

void	init_gfx(void);				/* initialize the graphics           */
void	draw_world(WorldDataType *wd, vrRenderInfo *rendinfo); /* render the world   */
void	init_world(WorldDataType *wd);		/* initialize the world parameters   */
void	update_world(WorldDataType *wd);	/* perform the world simulation      */


/***********************************************************************/
main(int argc, char* argv[])
{
	WorldDataType	*wd;

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, SIG_IGN);
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */
	vrStart();
	signal(SIGINT, exit_application);

	/********************/
	/* create the world */
	wd = (WorldDataType *)vrShmemAlloc0(sizeof(WorldDataType));
	init_world(wd);

	/***************************************/
	/* set up the FreeVR library functions */
	vrFunctionSetCallback(VRFUNC_DISPLAY_INIT, vrCallbackCreate(init_gfx, 0));
	vrFunctionSetCallback(VRFUNC_DISPLAY, vrCallbackCreate(draw_world, 1, wd));

	/********************************************************/
	/* define the application name, authors, controls, etc. */
	vrSystemSetName("FreeVR tutorial: example 7 -- pointer");
	vrSystemSetAuthors("Bill Sherman");
	vrSystemSetExtraInfo("An example to show a pointer attached to the wand");
	vrSystemSetStatusDescription("Application running fine");
	vrInputSet2switchDescription(0, "Terminate the application");
	vrInputSet2switchDescription(1, "Raise the green pyramid");
	vrInputSet2switchDescription(2, "Reset the green pyramid");
	vrInputSet2switchDescription(3, "Lower the green pyramid");
	vrInputSetValuatorDescription(0, "Move the green pyramid in the Z-plane");
	vrInputSetValuatorDescription(1, "Move the green pyramid in the X-plane");
	vrInputSet6sensorDescription(WAND_SENSOR, "Move a long cyan pointer");

	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* run until terminate key is pressed (Escape by default) */
	while(!vrGet2switchValue(0)) {
		vrFrame();
		update_world(wd);
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int signal)
{
	vrExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/


/**************************************************************/
/* init_world(): create the initial conditions of the dynamic */
/*   simulation.                                              */
/**************************************************************/
void init_world(WorldDataType *wd)
{
	int	objnum = 0;

	/* object 1 - a blue cube */
	wd->object[objnum].shape = cube;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 255;
	wd->object[objnum].x = -2.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -5.0;
	objnum++;

	/* object 2 - a red pyramid */
	wd->object[objnum].shape = pyramid;
	wd->object[objnum].color[0] = 255;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x =  2.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -5.0;
	objnum++;

	/* object 3 - an upside-down green pyramid */
	wd->object[objnum].shape = ud_pyramid;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x =  0.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -5.0;
	objnum++;

	wd->number_of_objects = objnum;
}


/*********************************************************************/
/* update_world(): Basic world with simple interactions.             */
/*********************************************************************/
void update_world(WorldDataType *wd)
{
#define JS_EPSILON	0.125				/* the dead-zone before joystick movement is recognized */
#define MOVE_FACTOR	2.5				/* a movement scale factor to tune control of pyramid   */
static	vrTime		last_time = -1.0;		/* initially set to an invalid value as a flag */
	vrTime		sim_time = vrCurrentSimTime();	/* the current time of the simulation  */
	double		delta_time;			/* time since last update (in seconds) */
	double		joy_x, joy_y;			/* Joystick values */

	/*****************************************************/
	/** Determine delta time from last simulation frame **/
	/*****************************************************/
	if (last_time == -1.0)
		delta_time = 0.0;
	else	delta_time = sim_time - last_time;
	last_time = sim_time;				/* now that delta_time has been calculated, we won't use last_time until next time */

	/* skip the update if the delta isn't big enough */
	if (delta_time <= 0.0)	/* can also choose a non-zero epsilon */
		return;

	/****************************************/
	/* Move the objects with constant speed */

	/* object 1 travels in a horizontal circle at a fixed speed */
	wd->object[0].x = 3*sin(sim_time * 2.0);
	wd->object[0].z = 3*cos(sim_time * 2.0) - 5.0;

	/* object 2 travels in a verticle circle at a fixed speed */
	wd->object[1].x = 3*cos(sim_time * 2.0);
	wd->object[1].y = 3*sin(sim_time * 2.0) + 5.0;

	/**************************/
	/* User interface section */

	/* the first button moves the object up, when initially pressed */
	if (vrGet2switchDelta(1) == 1) {
		wd->object[2].y += 1.0;
	}

	/* the second button resets the object's location, when initially pressed */
	if (vrGet2switchDelta(2) == 1) {
		wd->object[2].x =  0.0;
		wd->object[2].y =  5.0;
		wd->object[2].z = -5.0;
	}

	/* the third button moves the object down, when initially pressed */
	if (vrGet2switchDelta(3) == 1) {
		wd->object[2].y -= 1.0;
	}

	/* store input values so they are consistent throughout update routine */
	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	/* now use those values to move the object */
	if (fabs(joy_x) > JS_EPSILON)
		wd->object[2].x += joy_x * delta_time * MOVE_FACTOR;				/* NOTE: epsilon is NOT subtracted out */

	if (fabs(joy_y) > JS_EPSILON)
		wd->object[2].z -= (joy_y-copysign(JS_EPSILON,joy_y)) * delta_time * MOVE_FACTOR;/* NOTE: epsilon IS subtracted out */

	/****************************************************/
	/* yield to allow other processes access to the CPU */
	vrSleep(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

/* functions from shapes.c */
void	draw_cube();
void	draw_pyramid();
void	draw_cube_outline();
void	draw_pyramid_outline();


/*********************************************************************/
/* init_gfx(): initialize general graphics properties (eg. lighting) */
/*********************************************************************/
void init_gfx(void)
{
	GLfloat	location_0[4] = { 0.0, 10.0, 10.0, 1.0 /* 0 = directional, 1 = location */},
		ambient_0[4] = { 0.2, 0.2, 0.2, 1.0 },
		diffuse_0[4] = { 0.8, 0.8, 0.8, 1.0 },
		specular_0[4] = { 0.2, 0.2, 0.2, 1.0 };

	/****************************************/
	/* set the render initialization values */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);

	/********************************************************/
	/* set the polygon shading/lighting/material parameters */
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0, GL_POSITION, location_0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular_0);
	glEnable(GL_LIGHT0);

	/* These two lines cause the vertex colors to be used (otherwise all is grey) */
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_NORMALIZE);	/* this is required when calling glScalef() w/ lighting */

	/***************************/
	/* set the line parameters */
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(2.0);
}


/*******************************************************************/
/* draw_world(): Now will render how ever many objects are in the  */
/*   world data structure.  The color, type (and maybe rotation)   */
/*   are all now stored for each object, in addition to location.  */
/*******************************************************************/
void draw_world(WorldDataType *wd, vrRenderInfo *rendinfo)
{
static	GLfloat		location_0[4] = { 0.0, 10.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
static	char		msg[128];			/* memory for text string messages */
	int		objnum;				/* object counter */

	/**************************************/
	/* clear the screen -- very important */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/********************************************************************************/
	/* update the light's location -- must be done independently for each rendering */
	glLightfv(GL_LIGHT0, GL_POSITION, location_0);

	/******************************************/
	/* draw a pointer emanating from the wand */
	glPushMatrix();
		glColor3ub(100, 255, 255);	/* cyan */
		vrRenderTransform6sensor(rendinfo, WAND_SENSOR);
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glScalef(0.1, 1.5, 0.1);
		glTranslatef(0.0, 1.0, 0.0);
		draw_pyramid_outline();
	glPopMatrix();

	/************************/
	/* draw all the objects */

	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {

		glColor3ubv(wd->object[objnum].color);
		glPushMatrix();
			glTranslatef(wd->object[objnum].x, wd->object[objnum].y, wd->object[objnum].z);
			switch (wd->object[objnum].shape) {
			case cube:
				draw_cube();
				break;
			case pyramid:
				draw_pyramid();
				break;
			case ud_pyramid:
				glScalef(1.0, -1.0, 1.0);
				draw_pyramid();
				break;
			}
		glPopMatrix();
	}

	/*************************************************/
	/* draw text indicating the location of object 3 */
	sprintf(msg, "Green pyramid is at %.2f %.2f %.2f", wd->object[2].x, wd->object[2].y, wd->object[2].z);

	glDisable(GL_LIGHTING);		/* make the text bright regardless of lights */
	glColor3ub(255, 255, 255);	/* white */
	glPushMatrix();
		glTranslatef(-4.0, 7.0, -5.0);	/* on the front screen */

		glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
		vrRenderText(rendinfo, msg);
	glPopMatrix();
	glEnable(GL_LIGHTING);
}

