/***********************************************************************/
/* FreeVR OpenGL tutorial example shader2: exN_shader2.c               */
/* Last tested with: FreeVR version 0.6a                               */
/* Last Modified: March 19, 2011                                       */
/* URL: http://freevr.org                                              */
/***********************************************************************/
/* This example code demonstrates a VR application program that        */
/*   ...                                                               */
/*                                                                     */
/*                                                                     */
/* The new features of this example include:                           */
/*   -  */
/*                                                                     */
/* Considerations:                                                     */
/*   -  */
/*                                                                     */
/* This application uses the files:                                    */
/*   - shapes.c  (Bill Sherman's simple OpenGL shapes)                 */
/*                                                                     */
/* Copyright 2011, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/***********************************************************************/
#include <stdio.h>	/* needed for printf(), etc. */
#include <stdlib.h>	/* needed for exit() & getenv() */
#include <signal.h>	/* needed for signal() and its arguments */
#include <unistd.h>	/* not sure why/whether this is needed */
#include <math.h>	/* needed for fabs() */

#include <GL/gl.h>	/* needed for all OpenGL functions & values */

#include "freevr.h"	/* needed for FreeVR functions & values */
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */

/* Define some values to adjust the compilation of this example */
#define MAX_OBJECTS	4


/* Define a structure for information about each individual object. */
typedef struct {
		float			x, y, z;		/* location of the object     */
		GLubyte			color[3];		/* color of the object        */
		enum {	cube,
			pyramid,
			ud_pyramid }	shape;			/* shape of the object        */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		quit_bit;			/* a flag to pass along the desire to terminate */
		int		number_of_objects;		/* number of objects in world */
		int		selected_object;		/* currently selected object  */
		ObjectType	object[MAX_OBJECTS];		/* array of objects in world  */
		vrLock		lock;				/* to avoid bad multiprocess accesses */
	} WorldDataType;


/*************************************************************************************/
/* These function declarations are very important -- they will be used as arguments. */
void	exit_application(int signal);		/* cleanly close down operations     */

void	init_gfx();				/* initialize the graphics           */
void	init_shaders();				/* initialize the GLSL shaders       */
void	draw_world(WorldDataType *wd, vrRenderInfo *rendinfo); /* render the world   */
void	init_world(WorldDataType *wd);		/* initialize the world parameters   */
void	update_world(WorldDataType *wd);	/* perform the world simulation      */

void	end_world();				/* close down the world operations   */
void	init_objects(WorldDataType *wd);	/* initalize the objects in the world*/


/***********************************************************************/
main(int argc, char* argv[])
{
	WorldDataType	*wd;

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, SIG_IGN);
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */
	vrStart();
	signal(SIGINT, exit_application);

	/********************/
	/* create the world */
	wd = (WorldDataType *)vrShmemAlloc0(sizeof(WorldDataType));
	wd->lock = vrLockCreate();
	init_world(wd);

	/***************************************/
	/* set up the FreeVR library functions */
	vrFunctionSetCallback(VRFUNC_DISPLAY_INIT, vrCallbackCreate(init_gfx, 0));
	vrFunctionSetCallback(VRFUNC_DISPLAY, vrCallbackCreate(draw_world, 1, wd));

	/********************************************************/
	/* define the application name, authors, controls, etc. */
	vrSystemSetName("example 13 -- socket");
	vrSystemSetAuthors("Bill Sherman");
	vrSystemSetExtraInfo("An example to demonstrate communicating with the virtual world via sockets");
	vrSystemSetStatusDescription("Application running fine");
	vrInputSet2switchDescription(0, "Terminate the application");
	vrInputSet2switchDescription(1, "Move the selected object");
	vrInputSet2switchDescription(3, "Reset the travel matrix (take us home)");
	vrInputSetValuatorDescription(0, "Rotate us in the world about the Y-axis");
	vrInputSetValuatorDescription(1, "Move us though the world in the direction we point");
	vrInputSet6sensorDescription(WAND_SENSOR, "Move a short cyan pointer to manipulate objects & choose travel dir");

	/***************************/
	/*** do world simulation ***/
	/***************************/
	vrUserTravelReset(VR_ALLUSERS);
	/* run until the user has indicated it's time not to */
	while(!wd->quit_bit) {
		vrFrame();
		update_world(wd);
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int signal)
{
	end_world();

	vrExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

static int	listen_sock = -1;	/* socket to receive commands */
					/*   listen_sock is not in wd        */
					/*   because it needs to be          */
					/*   available to exit_application() */
					/*   which can have no arguments.    */


/****************************************************************/
/* init_world(): create the initial conditions of the dynamic   */
/*   simulation.  Now with the addition of code to initialize a */
/*   socket communication interface.                            */
/*                                                              */
/*   The socket server port to be used can be set from the      */
/*   "VR_PORT" environment variable.  If it is set, then its    */
/*   value is used as the initial server socket port.           */
/*   Otherwise, the arbitrarily chosen port number 4000 is used.*/
/*                                                              */
/*   Since IRIX, and perhaps other Unicies have a "feature" in  */
/*   which a socket port isn't usable again until after 60      */
/*   seconds have passed since the connection was closed, if the*/
/*   initial socket isn't available, we try the next two ports  */
/*   in succession.  The selected port is reported to the tty.  */
/****************************************************************/
void init_world(WorldDataType *wd)
{
	int	inet_port;
	int	inet_port_max;
	char	host_machine[128];

	/*************************/
	/* set world state flags */
	vrLockWriteSet(wd->lock);
	wd->quit_bit = 0;
	vrLockWriteRelease(wd->lock);

	/***************************/
	/* setup the world content */
	init_objects(wd);

	/******************************/
	/* initialize the socket port */
	gethostname(host_machine, 128);

	/* loop through N possible internet ports to hopefully find one that works */
	if (getenv("VR_PORT") != NULL)
		inet_port = atoi(getenv("VR_PORT"));
	else	inet_port = 4000;
	for (inet_port_max = inet_port + 2; listen_sock < 0 && inet_port <= inet_port_max; inet_port++) {
		printf("Port %d -- ", inet_port);
		fflush(stdout);
		listen_sock = vrSocketCreateListen(&inet_port, 0);
	}
	inet_port--;
	if (listen_sock < 0)
		printf("WARNING: Unable to find an open socket port.\n");
	else	printf("\rConnect to '%s:%d' to send commands to the simulation\n", host_machine, inet_port);
}


/***************************************************************/
/* init_objects(): place objects into the world data structure */
/*   at some initial positions.                                */
/***************************************************************/
void init_objects(WorldDataType *wd)
{
	int	objnum = 0;

	vrLockWriteSet(wd->lock);

	/* object 1 - a blue cube */
	wd->object[objnum].shape = cube;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 255;
	wd->object[objnum].x =  0.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	/* object 2 - a red pyramid */
	wd->object[objnum].shape = pyramid;
	wd->object[objnum].color[0] = 255;
	wd->object[objnum].color[1] = 100;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x = -4.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	/* object 3 - an upside-down green pyramid */
	wd->object[objnum].shape = ud_pyramid;
	wd->object[objnum].color[0] = 100;
	wd->object[objnum].color[1] = 255;
	wd->object[objnum].color[2] = 100;
	wd->object[objnum].x =  4.0;
	wd->object[objnum].y =  5.0;
	wd->object[objnum].z = -4.0;
	objnum++;

	wd->number_of_objects = objnum;
	wd->selected_object = -1;

	/* sanity test */
	if (objnum > MAX_OBJECTS) {
		printf("DANGER: Number of objects (%d), exceeds maximum allowable (%d).  You may experience issues.\n", objnum, MAX_OBJECTS);
	}

	vrLockWriteRelease(wd->lock);
}


/*******************************************************/
/* end_world(): clean up anything opened by simulation */
/*******************************************************/
void end_world()
{
	vrSocketClose(listen_sock);
}


/*********************************************************************/
/* update_world(): In this version, we handle the user interactions. */
/*   Both selecting, and moving the selected object.                 */
/*********************************************************************/
void update_world(WorldDataType *wd)
{
#define JS_EPSILON	0.125				/* the dead-zone before joystick movement is recognized */
#define MOVE_FACTOR	2.5				/* a scale factor to tune travel movement               */
#define TURN_FACTOR	25.0				/* a scale factor to tune travel movement               */
#define SUB_JS_EPSILON(a)	((a)-copysign(JS_EPSILON,(a)))	/* a macro to remove the joystick epsilon range */
static	char	*help_msg = "Socket commands:\n"
				"\thelp -- this message\n"
				"\tquit -- quit the application\n"
				"\tclose -- close the connection\n"
				"\treset -- reset the application\n"
				"\tmove cube up -- translate the cube up 2 feet\n"
				"\tmove cube down -- translate the cube down 2 feet\n"
				"\twhere -- report the user's position in the world\n"
				"\twhere am i -- report the user's position in the world\n"
				"\tshow ActiveWalls -- report some CAVE config info\n"
		"";
static	int	cmd_socket = -1;	/* socket to get commands from */
static	char	buf[128];		/* socket read buffer          */
static	char	msg[128];		/* socket send buffer          */
	int	result;			/* result of reading a socket  */

static	vrTime		last_time = -1.0;		/* initially set to an invalid value as a flag */
	vrTime		sim_time = vrCurrentSimTime();	/* the current time of the simulation  */
static	vrPoint		last_wand_vw_locpnt;		/* VW location of the wand in the last update frame */
	vrPoint		wand_vw_locpnt;			/* current location of the wand (in VW coords) */
#define	last_wand_vw_loc last_wand_vw_locpnt.v
#define	wand_vw_location wand_vw_locpnt.v
	vrVector	wand_rw_pointvec;		/* current vector pointing out of the wand (in RW coords) */
	double		delta_time;			/* time since last update (in seconds) */
	double		delta_move;			/* distance an object should be moved */
	double		joy_x, joy_y;			/* Joystick values */
	int		selected_obj;			/* the currently selected object  */
	int		touch_obj = -1;			/* the currently touched object   */
	int		objnum;				/* for looping throuh the objects */
	ObjectType	*object;			/* pointer to object to be moved  */

	/*****************************************************/
	/** Determine delta time from last simulation frame **/
	/*****************************************************/
	if (last_time == -1.0)
		delta_time = 0.0;
	else	delta_time = sim_time - last_time;
	last_time = sim_time;				/* now that delta_time has been calculated, we won't use last_time until next time */

	/* skip the update if the delta isn't big enough */
	if (delta_time <= 0.0)	/* can also choose a non-zero epsilon */
		return;

	/**********************************/
	/** Handle Manipulation via wand **/
	/**********************************/

	/**********************************************************/
	/* check for contact between the wand and all the objects */
	vrLockReadSet(wd->lock);

	vrPointGetVWFromUser6sensor(&wand_vw_locpnt, 0, WAND_SENSOR);

	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {
		object = &(wd->object[objnum]);
		if ( (wand_vw_location[VR_X] < object->x + 1.0)
		  && (wand_vw_location[VR_X] > object->x - 1.0)
		  && (wand_vw_location[VR_Y] < object->y + 1.0)
		  && (wand_vw_location[VR_Y] > object->y - 1.0)
		  && (wand_vw_location[VR_Z] < object->z + 1.0)
		  && (wand_vw_location[VR_Z] > object->z - 1.0) ) {
			touch_obj = objnum;
		}
	}

	selected_obj = wd->selected_object;

	vrLockReadRelease(wd->lock);

	if (selected_obj != touch_obj) {
		/* NOTE: a local variable (touch_obj) is used because it */
		/*   decreases the time the lock must be held.           */
		vrLockWriteSet(wd->lock);
		wd->selected_object = touch_obj;
		vrLockWriteRelease(wd->lock);

		selected_obj = touch_obj;
	}

	/* when left wand button is pressed, the selected object is moved */
	if (vrGet2switchValue(1) && (selected_obj != -1)) {
		vrLockWriteSet(wd->lock);
		object = &(wd->object[selected_obj]);
		object->x += (wand_vw_location[VR_X] - last_wand_vw_loc[VR_X]);
		object->y += (wand_vw_location[VR_Y] - last_wand_vw_loc[VR_Y]);
		object->z += (wand_vw_location[VR_Z] - last_wand_vw_loc[VR_Z]);
		vrLockWriteRelease(wd->lock);
	}

	/******************************************************************/
	/* store the current wand location for next time through the loop */
	last_wand_vw_loc[VR_X] = wand_vw_location[VR_X];
	last_wand_vw_loc[VR_Y] = wand_vw_location[VR_Y];
	last_wand_vw_loc[VR_Z] = wand_vw_location[VR_Z];


	/****************************/
	/** Handle Travel via wand **/
	/****************************/

	/* pressing the right wand button resets us to the initial position */
	if (vrGet2switchValue(3)) {
		vrUserTravelReset(VR_ALLUSERS);
	}

	/* use wand joystick to fly through world */
	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	if (fabs(joy_x) > JS_EPSILON)
		vrUserTravelRotateId(VR_ALLUSERS, VR_Y, (delta_time * SUB_JS_EPSILON(joy_x) * -TURN_FACTOR));

	if (fabs(joy_y) > JS_EPSILON) {
		delta_move = delta_time * SUB_JS_EPSILON(joy_y) * MOVE_FACTOR;
		vrVectorGetRWFrom6sensorDir(&wand_rw_pointvec, WAND_SENSOR, VRDIR_FORE);
		vrUserTravelTranslate3d(VR_ALLUSERS,
			wand_rw_pointvec.v[VR_X] * delta_move,
			wand_rw_pointvec.v[VR_Y] * delta_move,
			wand_rw_pointvec.v[VR_Z] * delta_move);
	}


	/*******************************/
	/* Handle Commands from socket */

	/* if no connection (and a listen socket is open), see if someone is calling */
	if (cmd_socket < 0 && listen_sock >= 0) {
		/* cmd_socket will be -1 when no one is connected */
		cmd_socket = vrSocketAnswer(listen_sock);
		if (cmd_socket > 0) {
			printf("New client connected.\n");
			vrSocketSendMsg(cmd_socket, "Your wish is my command >");
		}
	}

	/* if a connection, see if client sent a command.  If so, act on it */
	if (cmd_socket > 0) {
		result = vrSocketReadMsgClean(cmd_socket, buf, 128);
		switch (result) {
		case -1:
			break;
		case  0:
			printf("Client stopped communicating.\n");
			close(cmd_socket);
			cmd_socket = -1;
			break;
		default:
			printf("socket cmd = '%s'\n", buf);
			/* parse command here -- very simple-minded parsing */
			if (strcmp(buf, "quit") == 0) {
				/* set the world for termination */
				wd->quit_bit = 1;
			} else
			if (strcmp(buf, "help") == 0 || strcmp(buf, "?") == 0) {
				/* give usage information */
				vrSocketSendMsg(cmd_socket, help_msg);
			} else
			if (strcmp(buf, "close") == 0) {
				/* close socket connection */
				printf("Client ending communication.\n");
				close(cmd_socket);
				cmd_socket = -1;
			} else
			if (strcmp(buf, "reset") == 0) {
				/* make the world as it was in the beginning */
				init_objects(wd);
				vrUserTravelReset(VR_ALLUSERS);
			} else
			if (strcmp(buf, "move cube up") == 0) {
				/* move the blue cube up 2.0 units */
				wd->object[0].y += 2.0;
			} else
			if (strcmp(buf, "move cube down") == 0) {
				/* move the blue cube down 2.0 units */
				wd->object[0].y -= 2.0;
			} else
			if (strcmp(buf, "where") == 0 || strcmp(buf, "where am i") == 0) {
				/* report the user's location */
				vrMatrix	*headmat;
				vrPoint		headloc;

				headmat = vrMatrixCreate();
				vrMatrixGetRWFromUserHead(headmat, 0);
				vrPointGetVWFromUserMatrix(&headloc, 0, headmat);
				sprintf(msg, "You are at (%f %f %f).\n", headloc.v[VR_X], headloc.v[VR_Y], headloc.v[VR_Z]);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			if (strcmp(buf, "show ActiveWalls") == 0) {
				/* report the type of graphics (VR) display */
				sprintf(msg, "Application has %d active walls.\n", vrContext->config->num_windows);
				vrSocketSendMsg(cmd_socket, msg);
			} else
			/* default: */ {
				vrSocketSendMsg(cmd_socket, "Unknown command.\n");
			}

			/* prompt for another command (if socket still open) */
			if (cmd_socket > 0) {
				vrSocketSendMsg(cmd_socket, "Your wish is my command >");
			}
		}
	}

	/*********************************/
	/* Handle Commands from keyboard */
	if (vrGet2switchValue(0)) {
		wd->quit_bit = 1;
	}

	/****************************************************/
	/* yield to allow other processes access to the CPU */
	vrSleep(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

/* functions from shapes.c */
void	draw_floor();
void	draw_cube();
void	draw_pyramid();
void	draw_cube_outline();
void	draw_pyramid_outline();


/******************************************************************/
/* init_shaders(): initialize the GLSL shaders that will be used. */
/******************************************************************/
void init_shaders()
{
	GLuint	vshader;	/* handle to the vertex shader */
	GLuint	fshader;	/* handle to the fragment shader */
	GLuint	program;	/* handle to the shader program */

	/*** initialize the GLSL shaders ***/
	char *vshader_string =
		"#version 110\n"
		"// a pass-through vertex shader\n"
		"void main()\n"
		"{\n"
		"	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
		"	gl_FrontColor = gl_Color;\n"
		"}\n";
	char *fshader_string =
		"#version 110\n"
		"// a color-altering fragment shader -- greyscale\n"
		"void main()\n{\n"
		"	float	luminosity;\n"
		"\n"
		"	luminosity = gl_Color.r * 0.30 + gl_Color.b * 0.59 + gl_Color.g * 0.11;\n"
		"	gl_FragColor = vec4(luminosity, luminosity, luminosity, gl_Color.a);\n"
		"}\n";

	vshader = glCreateShader(GL_VERTEX_SHADER);
	fshader = glCreateShader(GL_FRAGMENT_SHADER);
	program = glCreateProgram();

	/* assign the two shaders to the program */
	glAttachShader(program, vshader);
	glAttachShader(program, fshader);

	/* compile the two shaders */
	glShaderSource(vshader, 1, &vshader_string, NULL);
	glCompileShader(vshader);

	/* vshader error checks */
	char	log[256];
	int	log_return = 0;
	glGetShaderInfoLog(vshader, 256, &log_return, log);
	if (log_return) {
		printf("Vertex Shader output:\n%s-------------\n", log);
		vshader = 0;
	}

	glShaderSource(fshader, 1, &fshader_string, NULL);
	glCompileShader(fshader);

	/* fshader error checks */
	log_return = 0;
	glGetShaderInfoLog(fshader, 256, &log_return, log);
	if (log_return) {
		printf("Fragment Shader output:\n%s-------------\n", log);
		fshader = 0;
	}

	/* link the shaders into the program & check for logged errors */
	glLinkProgram(program);
	log_return = 0;
	glGetProgramInfoLog(program, 256, &log_return, log);
	if (log_return) {
		printf("Shader Program linker output:\n%s-------------\n", log);
		program = 0;
	}

	/* enable the shader for the rest of the appllication */
	glUseProgram(program);
}


/*********************************************************************/
/* init_gfx(): initialize general graphics properties (eg. lighting) */
/*********************************************************************/
void init_gfx()
{
	GLfloat	location_0[4] = { 0.0, 10.0, 10.0, 1.0 /* 0 = directional, 1 = location */},
		ambient_0[4] = { 0.2, 0.2, 0.2, 1.0 },
		diffuse_0[4] = { 0.8, 0.8, 0.8, 1.0 },
		specular_0[4] = { 0.2, 0.2, 0.2, 1.0 };

	/****************************************/
	/* set the render initialization values */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);

	/********************************************************/
	/* set the polygon shading/lighting/material parameters */
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0, GL_POSITION, location_0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular_0);
	glEnable(GL_LIGHT0);

	/* These two lines cause the vertex colors to be used (otherwise all is grey) */
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	glEnable(GL_NORMALIZE);	/* this is required when calling glScalef() w/ lighting */

	/***************************/
	/* set the line parameters */
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(2.0);

	/***************************/
	/* initialize GLSL shaders */
	init_shaders();
}


/*******************************************************************/
/* draw_world(): Now will render how ever many objects are in the  */
/*   world data structure.  The color, type (and maybe rotation)   */
/*   are all now stored for each object, in addition to location.  */
/*   The currently selected object is highlighted by adding a      */
/*   wireframe outline to the representation.                      */
/*******************************************************************/
void draw_world(WorldDataType *wd, vrRenderInfo *rendinfo)
{
static	GLfloat		location_0[4] = { 0.0, 10.0, 0.0, 1.0 /* 0 = directional, 1 = location */};
static	char		msg[128];			/* memory for text string messages */
	int		objnum;				/* object counter */
	ObjectType	*object;			/* pointer to current object */

	/**************************************/
	/* clear the screen -- very important */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/********************************************************************************/
	/* update the light's location -- must be done independently for each rendering */
	glLightfv(GL_LIGHT0, GL_POSITION, location_0);

	/************************************/
	/* draw a short pointer at the wand */
	glPushMatrix();
		glColor3ub(100, 255, 255);	/* cyan */
		vrRenderTransform6sensor(rendinfo, WAND_SENSOR);
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glScalef(0.15, 0.15, 0.15);
		glTranslatef(0.0, 1.5, 0.0);
		draw_pyramid_outline();
	glPopMatrix();

	/*******************************************************/
	/* draw an object that remains fixed in physical space */
	glColor3ub(200, 100, 255);	/* purple */
	glPushMatrix();
		glTranslatef(-4.0, 1.0, -1.0);
		draw_cube();
	glPopMatrix();

	/******************************************************************/
	/* draw the rest of the objects in virtual (user navigated) space */
	glPushMatrix();		/* save the real-world coordinate system */
	vrRenderTransformUserTravel(rendinfo);	/* change to virtual-world coordinates */

	draw_floor();

	vrLockReadSet(wd->lock);	/* lock while reading info about objects */
	for (objnum = 0; objnum < wd->number_of_objects; objnum++) {
		object = &(wd->object[objnum]);

		glColor3ubv(object->color);
		glPushMatrix();
			glTranslatef(object->x, object->y, object->z);
			switch (object->shape) {
			case cube:
				if (objnum == wd->selected_object)
					draw_cube_outline();
				else	draw_cube();
				break;
			case pyramid:
				if (objnum == wd->selected_object)
					draw_pyramid_outline();
				else	draw_pyramid();
				break;
			case ud_pyramid:
				glScalef(1.0, -1.0, 1.0);
				if (objnum == wd->selected_object)
					draw_pyramid_outline();
				else	draw_pyramid();
				break;
			}
		glPopMatrix();
	}

	glPopMatrix();	/* return to real-world coordinates */
	/* NOTE: 2-D text is best rendered in a fixed location in the world */

	/*************************************************/
	/* draw text indicating which object is selected */
	if (wd->selected_object < 0) {
		sprintf(msg, "Object selected: none");
	} else {
		sprintf(msg, "Object selected: %d", wd->selected_object);
	}

	vrLockReadRelease(wd->lock);

	glDisable(GL_LIGHTING);		/* make the text bright regardless of lights */
	glColor3ub(255, 255, 255);	/* white */
	glPushMatrix();
		glTranslatef(-4.0, 7.0, -5.0);	/* on the front screen */

		glRasterPos3f(0.0, 0.0, 0.0);	/* go to the 3D location */
		vrRenderText(rendinfo, msg);
	glPopMatrix();
	glEnable(GL_LIGHTING);
}

