/****************************************************************************/
/**                                                                        **/
/** shapes.c                                                               **/
/**                                                                        **/
/** This file contains routines to draw simple geometric shapes.           **/
/**                                                                        **/
/**                                                                        **/
/* Copyright 2009, William R. Sherman, All Rights Reserved.            */
/*   In particular, the right to republish this file in whole or in    */
/*   part, in any form (including electronic forms) is prohibited      */
/*   without written consent of the copyright holder.                  */
/****************************************************************************/
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>


void	draw_cube();
void	draw_cube_outline();
void	draw_pyramid();
void	draw_pyramid_outline();
void	draw_floor();


/*****************************************************************/
/*
   Cube (and pyramid) vertices.  Labeled 0-7 (cube) & 8 (top of pyramid).


             7_______ 4
             /|     /|        (+y)
            / | .8 / |         |
          0/__|___/3 |         |
           | 6|___|__|5        o---- (+x)
           | /    | /         /
           |/     |/         /
          1|______|2       (+z)

*/



float cube_vertices[9][3] = {			/* specifies the (x,y,z)   */
		{ -1.0,  1.0,  1.0 },	/* 0 */	/* location of each vertex */
		{ -1.0, -1.0,  1.0 },	/* 1 */
		{  1.0, -1.0,  1.0 },	/* 2 */
		{  1.0,  1.0,  1.0 },	/* 3 */
		{  1.0,  1.0, -1.0 },	/* 4 */
		{  1.0, -1.0, -1.0 },	/* 5 */
		{ -1.0, -1.0, -1.0 },	/* 6 */
		{ -1.0,  1.0, -1.0 },	/* 7 */
		{  0.0,  1.0,  0.0 },	/* 8 */
	};

int cube_polygons[6][4] = {			/* specifices the vertices */
		{ 0, 1, 2, 3 },	/* front  */	/* of each polygon         */
		{ 3, 2, 5, 4 },	/* right  */
		{ 5, 6, 7, 4 },	/* back   */
		{ 6, 1, 0, 7 },	/* left   */
		{ 5, 2, 1, 6 },	/* bottom */
		{ 0, 3, 4, 7 },	/* top    */
	};

float cube_normals[6][3] = {			/* specifies the normal */
		{  0.0,  0.0,  1.0 }, /* front  ** of each polygon      */
		{  1.0,  0.0,  0.0 }, /* right  */
		{  0.0,  0.0, -1.0 }, /* back   */
		{ -1.0,  0.0,  0.0 }, /* left   */
		{  0.0, -1.0,  0.0 }, /* bottom */
		{  0.0,  1.0,  0.0 }, /* top    */
	};

int pyramid_polygons[5][4] = {			/* specifices the vertices */
		{ 1, 2, 8 },	/* front  */	/* of each polygon         */
		{ 2, 5, 8 },	/* right  */
		{ 5, 6, 8 },	/* back   */
		{ 6, 1, 8 },	/* left   */
		{ 6, 5, 2, 1 },	/* bottom */
	};

float pyramid_normals[5][3] = {			/* specifies the normal    */
		{  0.0,  0.0,  1.0 }, /* front  ** of each polygon -- only */
		{  1.0,  0.0,  0.0 }, /* right  ** the bottom normal is    */
		{  0.0,  0.0, -1.0 }, /* back   ** perpendicular to the    */
		{ -1.0,  0.0,  0.0 }, /* left   ** surface.                */
		{  0.0, -1.0,  0.0 }, /* bottom */
	};


/*****************************************************************/
void draw_cube()
{
	int	poly;

	for (poly = 0; poly < 6; poly++) {
		glBegin(GL_POLYGON);
			glNormal3fv(cube_normals[poly]);
			glVertex3fv(cube_vertices[cube_polygons[poly][0]]);
			glVertex3fv(cube_vertices[cube_polygons[poly][1]]);
			glVertex3fv(cube_vertices[cube_polygons[poly][2]]);
			glVertex3fv(cube_vertices[cube_polygons[poly][3]]);
		glEnd();
	}
}


/*****************************************************************/
void draw_cube_outline()
{
	draw_cube();
	glColor3ub(0, 0, 0);
	glBegin(GL_LINE_STRIP);
		glVertex3fv(cube_vertices[0]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[3]);
		glVertex3fv(cube_vertices[0]);
		glVertex3fv(cube_vertices[7]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[7]);
		glVertex3fv(cube_vertices[4]);
		glVertex3fv(cube_vertices[3]);
		glVertex3fv(cube_vertices[4]);
		glVertex3fv(cube_vertices[5]);
	glEnd();
}


/*****************************************************************/
void draw_pyramid()
{
	int	poly;

	for (poly = 0; poly < 5; poly++) {
		glBegin(GL_POLYGON);
			glNormal3fv(pyramid_normals[poly]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][0]]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][1]]);
			glVertex3fv(cube_vertices[pyramid_polygons[poly][2]]);
			if (poly == 4)
				glVertex3fv(cube_vertices[pyramid_polygons[poly][3]]);
		glEnd();
	}
}


/*****************************************************************/
void draw_pyramid_outline()
{
	draw_pyramid();
	glColor3ub(0, 0, 0);
	glBegin(GL_LINE_STRIP);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[8]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[1]);
		glVertex3fv(cube_vertices[6]);
		glVertex3fv(cube_vertices[8]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[2]);
		glVertex3fv(cube_vertices[5]);
		glVertex3fv(cube_vertices[6]);
	glEnd();
}


/*****************************************************************/
/* draw_floor(): draws a floor especially designed for the CAVE. */
/*****************************************************************/
void draw_floor()
{
	glColor3ub(25, 25, 200);
	glBegin(GL_POLYGON);
		glNormal3f( 0.0, 1.0,  0.0);
		glVertex3f(-5.0, 0.0, -5.0);
		glVertex3f( 5.0, 0.0, -5.0);
		glVertex3f( 5.0, 0.0,  5.0);
		glVertex3f(-5.0, 0.0,  5.0);
	glEnd();
}


